class Knob:
    @classmethod
    def from_dict(cls, data):
        return cls(data["name"], data["value"], data["weights"])

    def __init__(self, name, value=0, weights=None):
        self.name = name
        self.value = value
        self.weights = weights

    def from_madx(self, madx, redefine_weights=False):
        if weights is None:
            weights = {}

    def __repr__(self):
        return f"Knob({self.name!r}, {self.value})"

    def to_dict(self):
        return {
            "name": self.name,
            "value": self.value,
            "weights": self.weights,
        }



import numpy as np
from scipy.interpolate import interp1d

class Piecewise:
    def __init__(self, t, v):
        self.t = t
        self.v = v

    def __call__(self, t):
        return np.interp(t, self.t, self.v)

class Polynomial:
    @classmethod
    def from_fit(cls, t, v, n, t0=None, v0=None, t1=None, v1=None, t2=None, v2=None):
        """Fit best polynomial of degree n to data points (t, v) and optional fixed point (t0, v0) and derivative (t1, v1) and second derivative (t2, v2)
        """
        coeffs = np.polyfit(t, v, n)
        return cls(coeffs)

    def __init__(self, coeffs):
        self.coeffs = coeffs

    def __call__(self, x):
        return sum(c * x ** i for i, c in enumerate(self.coeffs))

class Spline:
    @classmethod
    def from_fit(cls, t, v):
        spline=interp1d(t, v, kind="cubic")
        return cls(spline)

    def __init__(self, spline):
        self.spline = spline

    def __call__(self, t):
        self.spline(t)


import re
from .knob import Knob

import pyoptics


class LHCMadModel:
    extra_madx = """
kd1.lr1       :=  ad1.lr1/l.mbxw;
kd2.l1        :=  ad2.l5/l.mbrc ;
kd2.r1        :=  ad2.r5/l.mbrc ;
kd1.l2        :=  ad1.l2/l.mbx  ;
kd1.r2        :=  ad1.r2/l.mbx  ;
kd2.l2        :=  ad2.l2/l.mbrc ;
kd2.r2        :=  ad2.r2/l.mbrc ;
kd3.lr3       :=  ad3.lr3/l.mbw ;
kd4.lr3       :=  ad4.lr3/l.mbw ;
kd3.l4        :=  ad3.l4/l.mbrs ;
kd3.r4        :=  ad3.r4/l.mbrs ;
kd4.l4        :=  ad4.l4/l.mbrb ;
kd4.r4        :=  ad4.r4/l.mbrb ;
kd34.lr3      :=  ad3.lr3/l.mbw ;
kd34.lr7      :=  ad3.lr7/l.mbw ;
kd1.lr5       :=  ad1.lr5/l.mbxw;
kd2.l5        :=  ad2.l5/l.mbrc ;
kd2.r5        :=  ad2.r5/l.mbrc ;
kd3.lr7       :=  ad3.lr7/l.mbw ;
kd4.lr7       :=  ad4.lr7/l.mbw ;
kd1.l8        :=  ad1.l8/l.mbx  ;
kd1.r8        :=  ad1.r8/l.mbx  ;
kd2.l8        :=  ad2.l8/l.mbrc ;
kd2.r8        :=  ad2.r8/l.mbrc ;
ksumd2.l1b2   :=  kd2.l1        ;
ksumd2.l2b2   :=  kd2.l2        ;
ksumd2.l5b2   :=  kd2.l5        ;
ksumd2.l8b2   :=  kd2.l8        ;
ksumd2.r1b2   :=  kd2.l1        ;
ksumd2.r2b2   :=  kd2.l2        ;
ksumd2.r5b2   :=  kd2.l5        ;
ksumd2.r8b2   :=  kd2.l8        ;

kb.a12        :=  ab.a12/l.mb   ;
kb.a23        :=  ab.a23/l.mb   ;
kb.a34        :=  ab.a34/l.mb   ;
kb.a45        :=  ab.a45/l.mb   ;
kb.a56        :=  ab.a56/l.mb   ;
kb.a67        :=  ab.a67/l.mb   ;
kb.a78        :=  ab.a78/l.mb   ;
kb.a81        :=  ab.a81/l.mb   ;

abas:= 12.00/ 6.0*clight/(7E12)*on_sol_atlas;
abls:= 6.05/12.1*clight/(7E12)*on_sol_alice ;
abcs:= 52.00/13.0*clight/(7E12)*on_sol_cms  ;
abxwt.l2      := -0.0000772587268993839836*on_alice ;
abwmd.l2      := +0.0001472587268993839840*on_alice ;
abaw.r2       := -0.0001335474860334838000*on_alice ;
abxwt.r2      := +0.0000635474860334838004*on_alice ;
abxws.l8      := -0.000045681598453109894*on_lhcb   ;
abxwh.l8      := +0.000180681598453109894*on_lhcb   ;
ablw.r8       := -0.000180681598453109894*on_lhcb   ;
abxws.r8      := +0.000045681598453109894*on_lhcb   ;


beam, particle=proton, energy=6.8e12, sequence=lhcb1;
beam, particle=proton, energy=6.8e12, sequence=lhcb2, bv=-1;
use, sequence=lhcb1;
use, sequence=lhcb2;
    """

    @classmethod
    def from_madxfile(cls, madxfile):
        from cpymad.madx import Madx

        madx = Madx()
        madx.options(echo=False, warn=False, info=False)
        madx.call(str(madxfile))
        madx.input(cls.extra_madx)
        return cls(madx)

    def __init__(self, madx):
        self.madx = madx
        self.b1 = MADSequence(self.madx, "lhcb1")
        self.b2 = MADSequence(self.madx, "lhcb2")

    def __getitem__(self, key):
        return self.madx.globals[key]

    def __setitem__(self, key, value):
        self.madx.globals[key] = value

    def filter(self, pattern):
        return [k for k in self.madx.globals if re.match(pattern, k)]

    def update_vars(self, strengths):
        for k, v in strengths.items():
            self[k] = v

    def update_knobs(self, knobs):
        for k, knob in knobs.items():
            self[k] = knob.value
            for wn, value in knob.weights.items():
                name = f"{wn}_from_{k}"
                self[name] = value
                if expr := self.madx.globals.cmdpar[wn].expr:
                    if expr is not None and k in expr and name not in expr:
                        raise ValueError(
                            f"{wn} depends on {k} but not on {name}"
                        )
                if expr is None or (expr is not None and k not in expr):
                    self.madx.input(f"{wn} = expr + ({name} * {k});")

    def update(self, src):
        if hasattr(src, "strengths"):
            self.update_vars(src.strengths)
        else:
            self.update_vars(src)
        if hasattr(src, "knobs"):
            self.update_knobs(src.knobs)

    def mad_mkknob(self, name):
        madx=self.madx
        base = dict(madx.globals)
        value = madx.globals.get(name, 0)
        dvalue = value + 1
        weights = {}
        for s in base:
            dvalue = madx.globals[s] - base[s]
            if dvalue != 0:
                weights[s] = dvalue
        madx.globals[name] = value
        return Knob(name, value, weights)


    def mad_find_and_set0_knobs(madx, strengths):
        """
        Find knobs that are used to set the strengths and set knob value to 0
        """
        defs = {}
        for st in strengths:
            if expr := madx.globals.cmdpar[st].expr:
                for knob in re.findall("on[A-z_0-9.]*", expr):
                    if len(knob) < 12:
                        defs.setdefault(knob, []).append(st)
        knobs = {}
        for knob, strengths in defs.items():
            value = madx.globals[knob]
            madx.globals[knob] = 0
            base = {st: madx.globals[st] for st in strengths}
            weights = {}
            madx.globals[knob] = 1
            for s in strengths:
                dvalue = madx.globals[s] - base[s]
                weights[s] = dvalue
            madx.globals[knob] = 0
            knobs[knob] = Knob(knob, value, weights)
        return knobs


    def make_and_set0_knobs(self, knob_names):
        madx=self.madx
        defs = {}
        for knob in knob_names:
            for st in madx.globals:
                if expr := madx.globals.cmdpar[st].expr:
                    if knob in expr:
                        defs.setdefault(knob, []).append(st)
        knobs = {}
        for knob, strengths in defs.items():
            value = madx.globals[knob]
            madx.globals[knob] = 0
            base = dict(madx.globals)
            weights = {}
            madx.globals[knob] = 1
            for s in strengths:
                dvalue = madx.globals[s] - base[s]
                weights[s] = dvalue
            madx.globals[knob] = 0
            knobs[knob] = Knob(knob, value, weights)
        return knobs

class MADSequence:
    def __init__(self, madx, sequence_name):
        self.madx = madx
        self.sequence = sequence_name

    @property
    def madx_sequence(self):
        return self.madx.sequence[self.sequence]

    @property
    def start(self):
        self.madx_sequence.elemnts[1].name

    @property
    def end(self):
        self.madx_sequence.elemnts[-2].name

    def twiss(
        self,
        start=None,
        end=None,
        betx=None,
        bety=None,
        alfx=None,
        alfy=None,
        dx=None,
        dpx=None,
        dy=None,
        dpy=None,
        init=None,
        s=0,
        at=None,
        full=True,
    ):
        """
        Parameters:
        - full: if False consider only the portion from start to end
        - at: location to which initial conditions are to be set
        - init, betx, .... : initial conditions if all are None, then they are taken from the periodic solution at `at`
        - start, end: start and end of the twiss output
        """
        if set([betx, bety, alfx, alfy, dx, dpx, dy, dpy, init]) == {
            None
        }:  # is periodic
            self.madx.use(sequence=self.sequence, range=f"{start}/{end}")
            tw = self.madx.twiss(sequence=self.sequence)
        if full is False:  # consider only the machine from start to end
            if start is None:
                start = self.madx.sequence
            if end is None:
                end = self.madx.globals[self.sequence].sequence.end
        else:
            tw = self.madx.twiss(
                betx=betx,
                bety=bety,
                alfx=alfx,
                alfy=alfy,
                sequence=self.sequence,
            )
        return pyoptics.optics(tw)

    def twiss_periodic(self, use=False):
        if use:
            self.madx.use(sequence=self.sequence)
        tw = self.madx.twiss(sequence=self.sequence)
        return pyoptics.optics(tw)

    def twiss_line(self, betx=1, bety=1, alfx=0, alfy=0):
        tw = self.madx.twiss(
            betx=betx, bety=bety, alfx=alfx, alfy=alfy, sequence=self.sequence
        )
        return pyoptics.optics(tw)

    def rmatrix(self, start=None, end=None):
        if start is None:
            start = self.start
        if end is None:
            end = self.end
        self.madx.use(sequence=self.sequence, range=f"{start}/{end}")
        tw = self.madx.twiss(betx=1, bety=1, rmatrix=True)
        return RMatrix(tw.getmat("re", -1, 6, 6))

    def twiss_line_back(
        self, start, end, betx=1, bety=1, alfx=0, alfy=0, dx=0, dpx=0
    ):
        self.madx.use(sequence=self.sequence, range=f"{start}/{end}")
        tw = self.madx.twiss(betx=1, bety=1, rmatrix=True)
        rmat = tw.getmat("re", -1, 6, 6)
        r11 = rmat[0, 0]
        r12 = rmat[0, 1]
        r21 = rmat[1, 0]
        r22 = rmat[1, 1]
        r33 = rmat[2, 2]
        r34 = rmat[2, 3]
        r43 = rmat[3, 2]
        r44 = rmat[3, 3]
        r16 = rmat[0, 5]
        r26 = rmat[1, 5]
        bx2 = betx
        by2 = bety
        ax2 = alfx
        ay2 = alfy
        dx2 = dx
        dpx2 = dpx
        gx2 = (1 + ax2**2) / bx2
        gy2 = (1 + ay2**2) / by2
        bx1 = r22**2 * bx2 + 2 * r22 * r12 * ax2 + r12**2 * gx2
        ax1 = r22 * r21 * bx2 + (r11 * r22 + r12 * r21) * ax2 + r12 * r11 * gx2
        by1 = r44**2 * by2 + 2 * r44 * r34 * ay2 + r34**2 * gy2
        ay1 = r44 * r43 * by2 + (r33 * r44 + r34 * r43) * ay2 + r34 * r33 * gy2
        dx1 = r22 * (dx2 - r16) - r12 * (dpx2 - r26)
        dpx1 = -r21 * (dx2 - r16) + r11 * (dpx2 - r26)
        tw = self.madx.twiss(
            betx=bx1, bety=by1, alfx=ax1, alfy=ay1, dx=dx1, dpx=dpx1
        )
        return tw


class RMatrix:
    @classmethod
    def from_twiss(cls, tw):
        pass

    def __init__(self, matrix):
        self.matrix = matrix


class TwissInit:
    def __init__(
        self,
        betx=1,
        alfx=1,
        mux=0,
        bety=1,
        alfy=0,
        muy=0,
        x=0,
        px=0,
        y=0,
        py=0,
        t=0,
        pt=0,
        dx=0,
        dpx=0,
        dy=0,
        dpy=0,
        wx=0,
        phix=0,
        dmux=0,
        wy=0,
        phiy=0,
        dmuy=0,
        ddx=0,
        ddpx=0,
        ddy=0,
        ddpy=0,
        r11=0,
        r12=0,
        r21=0,
        r22=0,
    ):
        self.betx = betx
        self.alfx = alfx
        self.mux = mux
        self.bety = bety
        self.alfy = alfy
        self.muy = muy
        self.x = x
        self.px = px
        self.y = y
        self.py = py
        self.t = t
        self.pt = pt
        self.dx = dx
        self.dpx = dpx
        self.dy = dy
        self.dpy = dpy
        self.wx = wx
        self.phix = phix
        self.dmux = dmux
        self.wy = wy
        self.phiy = phiy
        self.dmuy = dmuy
        self.ddx = ddx
        self.ddpx = ddpx
        self.ddy = ddy
        self.ddpy = ddpy
        self.r11 = r11
        self.r12 = r12
        self.r21 = r21
        self.r22 = r22


from .irs import LHCIR


class LHCIR1(LHCIR):
    name = "ir1"
    knobs = [
        "on_x1_h",
        "on_sep1_h",
        "on_x1_v",
        "on_sep1_v",
        "on_xip1b1",
        "on_xip1b2",
        "on_oh1",
        "on_yip1b2",
        "on_yip1b1",
        "on_ov1",
    ]



from .irs import LHCIR


class LHCIR5(LHCIR):
    name = "ir5"
    knobs = [
        "on_x5_h",
        "on_sep5_h",
        "on_x5_v",
        "on_sep5_v",
        "on_xip5b1",
        "on_xip5b2",
        "on_oh5",
        "on_yip5b2",
        "on_yip5b1",
        "on_ov5",
    ]



from cpymad.madx import Madx
import json
from pathlib import Path

from .ir1 import LHCIR1
from .ir2 import LHCIR2
from .ir3 import LHCIR3
from .ir4 import LHCIR4
from .ir5 import LHCIR5
from .ir6 import LHCIR6
from .ir7 import LHCIR7
from .ir8 import LHCIR8
from .arcs import LHCArc

from .section import Knob
from .model_xsuite import LHCXsuiteModel, LHCMadModel

irs = [LHCIR1, LHCIR2]

_opl = ["_op", "_sq", ""]


class LHCOptics:
    """
    Optics containts global knobs, global parameters and sections
    Section contains strengths, local knobs, local parameters
    """

    @staticmethod
    def set_repository(version="2024"):
        import subprocess
        import os
        accmodels = Path("acc-models-lhc")
        if accmodels.exists():
            if not (accmodels/"lhc.seq").exists():
                raise FileNotFoundError("acc-models-lhc/lhc.seq not found")
            else:
                if  (accmodels/".git").exists():
                    subprocess.run(["git", "switch", version], cwd=accmodels)
        elif (lcl:=(Path.home()/"local"/"acc-models-lhc"/version)).exists():
            accmodels.symlink_to(lcl)
        else:
            subprocess.run(["git", "clone", "https://gitlab.cern.ch/acc-models/lhc.git", "acc-models-lhc"])


    irs = [LHCIR1, LHCIR2, LHCIR3, LHCIR4, LHCIR5, LHCIR6, LHCIR7, LHCIR8]
    arcs = ["a12", "a23", "a34", "a45", "a56", "a67", "a78", "a81"]

    knobs = [f"dq{x}.b{b}{op}" for x in "xy" for b in "12" for op in _opl]
    knobs += [f"dqp{x}.b{b}{op}" for x in "xy" for b in "12" for op in _opl]
    knobs += [f"cm{x}s.b{b}{op}" for x in "ir" for b in "12" for op in _opl]
    knobs += [
        f"{kk}.b{b}"
        for kk in ["on_mo", "phase_change", "dp_trim"]
        for b in "12"
    ]
    knobs += ["on_ssep1_h", "on_xx1_v", "on_ssep5_v", "on_xx5_h"]

    def __init__(self, name, irs, arcs, params=None, knobs=None, model=None):
        self.name = name
        for ir in irs:
            setattr(self, ir.name, ir)
            ir.parent = self
        for arc in arcs:
            setattr(self, arc.name, arc)
            arc.parent = self
        self._irs = irs
        self._arcs = arcs
        if params is None:
            params = {}
        if knobs is None:
            knobs = {}
        self.params = params
        self.knobs = knobs
        self.model = model

    @classmethod
    def from_madxfile(
        cls, filename, name="lhcoptics", sliced=False, model=None
    ):
        madx = Madx()
        madx.call(filename)
        return cls.from_madx(madx, name=name, sliced=sliced, model=model)

    @classmethod
    def from_madx(cls, madx, name="lhcoptics", sliced=False, model=None):
        madmodel = LHCMadModel(madx)
        knobs = madmodel.make_and_set0_knobs(cls.knobs)
        irs = [ir.from_madx(madx) for ir in cls.irs]
        arcs = [LHCArc.from_madx(madx, arc) for arc in cls.arcs]
        for k, knob in knobs.items():
            madx.globals[k] = knob.value
        self = cls(name, irs, arcs, knobs=knobs)
        if model=='xsuite':
            model = LHCXsuiteModel.from_madx(madx,sliced=sliced)
            self.model=model
        elif model=='madx':
            self.model = madmodel
        return self


    @classmethod
    def from_dict(cls, data):
        irs = [
            globals()[f"LHCIR{n+1}"].from_dict(d)
            for n, d in enumerate(data["irs"])
        ]
        arcs = [LHCArc.from_dict(d) for d in data["arcs"]]
        return cls(
            name=data["name"],
            irs=irs,
            arcs=arcs,
            params=data["params"],
            knobs={k: Knob.from_dict(d) for k, d in data["knobs"].items()},
        )

    @classmethod
    def from_json(cls, filename):
        with open(filename) as f:
            data = json.load(f)
            return cls.from_dict(data)

    def __repr__(self) -> str:
        return f"<LHCOptics {self.name!r}>"

    def to_dict(self):
        return {
            "name": self.name,
            "irs": [ir.to_dict() for ir in self._irs],
            "arcs": [arc.to_dict() for arc in self._arcs],
            "params": self.params,
            "knobs": {n: k.to_dict() for n, k in self.knobs.items()},
        }

    def copy(self):
        return self.__class__.from_dict(self.to_dict())

    def to_json(self, filename):
        with open(filename, "w") as f:
            json.dump(self.to_dict(), f, indent=2)

    def update_model(self):
        for ss in self._irs + self._arcs:
            ss.update_model()
        return self

    def update_from_model(self):
        for ss in self._irs + self._arcs:
            ss.update_from_model()
        return self

    def set_xsuite_model(self, model):
        if Path(model).exists():
            model = LHCXsuiteModel.from_json(model)
        self.model = model
        self.update_model()
        return self

    def set_madx_model(self, model):
        if Path(model).exists():
            model = LHCMadModel.from_madxfile(model)
        self.model = model
        self.update_model()
        return self

    def get_params(self):
        tw1 = self.model.b1.twiss(compute_chromatic_properties=True)
        tw2 = self.model.b2.twiss(compute_chromatic_properties=True)
        params = {
            "qxb1": tw1.mux[-1],
            "qyb1": tw1.muy[-1],
            "qxb2": tw2.mux[-1],
            "qyb2": tw2.muy[-1],
            "qpxb1": tw1.mux[-1],
            "qpyb1": tw1.muy[-1],
            "qpxb2": tw2.mux[-1],
            "qpyb2": tw2.muy[-1],
        }
        for ss in self._irs + self._arcs:
            params.update(ss.get_params_from_twiss(tw1, tw2))
        return params



import re
import json

from .model_madx import LHCMadModel
from .knob import Knob

_lb = [(l, b) for l in "lr" for b in "12"]
_ac = {
    0: [("h", "l", "1"), ("v", "l", "1"), ("h", "r", "2"), ("v", "r", "2")],
    1: [("h", "l", "2"), ("v", "l", "2"), ("h", "r", "1"), ("v", "r", "1")],
}


def lhcprev(n):
    return (n - 2) % 8 + 1


def lhcsucc(n):
    return n % 8 + 1

def sort_n(lst):
    out = []
    for s in lst:
        if m := re.match(r"[a-z]*([0-9][0-9]?)", s):
            out.append((int(m.group(1)), s))
        else:
            out.append((0, s))
    return [s for _, s in sorted(out)]


class LHCSection:
    """
    Model a section of the machine.
    It contains local strengths, local parameters and local knobs.

    It can twiss, plot and match the section using the model in the parent.

    It can be updated from the model and update the model with the local values.

    """
    def __init__(
        self,
        name,
        start,
        end,
        strengths=None,
        params=None,
        knobs=None,
        parent=None,
    ):
        self.name = name
        self.start = start
        self.end = end
        if strengths is None:
            strengths = {}
        if params is None:
            params = {}
        if knobs is None:
            knobs = {}
        self.strengths = strengths
        self.params = params
        self.knobs = knobs
        self.parent = parent

    @property
    def model(self):
        return self.parent.model

    @classmethod
    def from_dict(cls, data):
        return cls(
            name=data["name"],
            start=data["start"],
            end=data["end"],
            strengths=data["strengths"],
            params=data["params"],
            knobs={k: Knob.from_dict(d) for k, d in data["knobs"].items()},
        )

    @classmethod
    def from_json(cls, filename):
        with open(filename) as f:
            data = json.load(f)
            return cls.from_dict(data)

    @classmethod
    def from_madxfile(cls, filename):
        model=LHCMadModel.from_madxfile(filename)
        return cls.from_madx(model.madx)

    def update_from_madxfile(self,filename):
        self.__class__.from_madxfile(filename)
        self.update(self.__class__.from_madxfile(filename))


    def to_dict(self):
        return {
            "name": self.name,
            "start": self.start,
            "end": self.end,
            "strengths": self.strengths,
            "params": self.params,
            "knobs": {n: k.to_dict() for n, k in self.knobs.items()},
        }

    def to_json(self, filename):
        with open(filename, "w") as f:
            json.dump(self.to_dict(), f)
        return {
            "name": self.name,
            "start": self.start,
            "end": self.end,
            "strengths": self.strengths,
            "params": self.params,
            "knobs": self.knobs,
        }

    def __repr__(self):
        return f"<LHCSection {self.name} {self.start}/{self.end}>"

    def copy(self):
        return self.__class__.from_dict(self.to_dict())

    def update_model(self, src=None):
        if self.model is None:
            raise ValueError("Model not set")
        self.model.update_vars(self.strengths)
        self.model.update_knobs(self.knobs)
        if src is not None:
            self.model.update_vars(src)
        return self

    def update_strengths(self,src=None):
        """
        Update existing stregnths from self. model or src.strengths or src
        """
        if src is None:
            src = self.model
        elif hasattr(src,"strengths"):
            src = src.strengths
        for k in self.strengths:
                self.strengths[k] = src[k]

    def update_knobs(self,src=None):
        """
        Update existing knobs from self. model or src.knobs or src
        """
        if src is None:
            src = self.model
            raise NotImplementedError
        elif hasattr(src,"knobs"):
            src=src.knobs
        for k in self.knobs:
            self.knobs[k] = src[k]

    def update_params(self,src=None):
        """
        Update existing params from self. model or src.params or src
        """
        if src is None:
            src = self.model
        elif hasattr(src,"params"):
            src = src.params
        for k in self.params:
            self.params[k] = src[k]

    def update(self, src=None):
        """
        Update existing strengths, knobs, params from self. model or src.params or src
        """
        self.update_strengths(src)
        self.update_knobs(src)
        self.update_params(src)
        return self



from .section import LHCSection
from .model_xsuite import LHCMadModel
import xtrack as xt


class LHCArc(LHCSection):
    @classmethod
    def from_madx(cls, madx, name="a12"):
        madmodel = LHCMadModel(madx)
        i1, i2 = int(name[1]), int(name[2])
        strength_names = []
        strength_names += madmodel.filter(f"kq[fd]\.*{name}$")
        strength_names += madmodel.filter(f"kqt[fd]\.*{name}b[12]$")
        strength_names += madmodel.filter(f"kqs\.*{name}b[12]$")
        strength_names += madmodel.filter(f"ksq\.*r{i1}b[12]$")
        strength_names += madmodel.filter(f"ksq\.*l{i2}b[12]$")
        strength_names += madmodel.filter(f"ks[fd][12]\.*{name}b[12]$")
        strength_names += madmodel.filter(f"ko[fd]\.*{name}b[12]$")
        knobs = {}
        params = {}
        strengths = {st: madx.globals[st] for st in strength_names}
        return cls(name, strengths, params, knobs)


    def __init__(
        self,
        name=None,
        strengths=None,
        params=None,
        knobs=None,
        start=None,
        end=None,
    ):
        i1, i2 = int(name[1]), int(name[2])
        start = f"s.ds.r{i1}"
        end = f"e.ds.l{i2}"
        super().__init__(name, start, end, strengths, params, knobs)
        self.i1 = i1
        self.i2 = i2
        self.start_cell = {1: f"s.cell.{i1}{i2}.b1", 2: f"s.cell.{i1}{i2}.b2"}
        self.end_cell = {1: f"e.cell.{i1}{i2}.b1", 2: f"e.cell.{i1}{i2}.b2"}
        self.startb = {1: f"s.ds.r{i1}.b1", 2: f"s.ds.r{i1}.b2"}
        self.endb = {1: f"e.ds.l{i2}.b1", 2: f"e.ds.l{i2}.b2"}

    def twiss_init(self, beam):
        tw = self.twiss(beam)
        return [
            tw.get_twiss_init(self.startb[beam]),
            tw.get_twiss_init(self.endb[beam]),
        ]

    def twiss_cell(self, beam=None):
        if beam is None:
            return [self.twiss_cell(beam=1), self.twiss_cell(beam=2)]
        sequence = self.model.sequence[beam]
        start_cell = self.start_cell[beam]
        end_cell = self.end_cell[beam]
        return sequence.twiss(start=start_cell, end=end_cell, init="periodic")

    def twiss_cell_init(self, beam):
        sequence = self.model.sequence[beam]
        start_cell = self.start_cell[beam]
        end_cell = self.end_cell[beam]
        twinit_cell = sequence.twiss(
            start=start_cell,
            end=end_cell,
            init="periodic",
            only_twiss_init=True,
        )
        return twinit_cell

    def twiss_full(self, beam=None):
        if beam is None:
            return [self.twiss_full(beam=1), self.twiss_full(beam=2)]
        else:
            sequence = self.model.sequence[beam]
            start = self.startb12[beam]
            end = self.endb12[beam]
            init = sequence.twiss().get_twiss_init(start)
            return sequence.twiss(start=start, end=end, init=init)

    def twiss(self, beam=None):
        if beam is None:
            return [self.twiss(beam=1), self.twiss(beam=2)]
        else:
            twinit_cell = self.twiss_cell_init(beam)
            start_arc = self.startb[beam]
            end_arc = self.endb[beam]

            sequence = self.model.sequence[beam]
            res = sequence.twiss(
                start=start_arc, end=end_arc, init=twinit_cell
            )

            res["mux"] = res["mux"] - res["mux", start_arc]
            res["muy"] = res["muy"] - res["muy", start_arc]

            return res

    def plot(self, beam=None, method="periodic", figlabel=None):
        if beam is None:
            return [self.plot(beam=1), self.plot(beam=2)]
        else:
            twiss = self.twiss(beam)
            if figlabel is None:
                figlabel = f"{self.name}b{beam}"
            return twiss.plot(figlabel=figlabel)

    def get_params_from_twiss(self, tw1, tw2):
        params = {
            f"mux{self.name}b1": tw1.mux[-1],
            f"muy{self.name}b1": tw1.muy[-1],
            f"mux{self.name}b2": tw2.mux[-1],
            f"muy{self.name}b2": tw2.muy[-1],
            f"muxcell{self.name[1:]}b1": tw1["mux", self.end_cell[1]]
            - tw1["mux", self.start_cell[1]],
            f"muxcell{self.name[1:]}b2": tw2["mux", self.end_cell[2]]
            - tw2["mux", self.start_cell[2]],
            f"muycell{self.name[1:]}b1": tw1["muy", self.end_cell[1]]
            - tw1["muy", self.start_cell[1]],
            f"muycell{self.name[1:]}b2": tw2["muy", self.end_cell[2]]
            - tw2["muy", self.start_cell[2]],
        }
        return params


    def get_params(self):
        """Get params from model"""
        tw1, tw2 = self.twiss()
        return self.get_params_from_twiss(tw1, tw2)


class ActionArcPhaseAdvance(xt.Action):
    def __init__(self, arc, beam):
        self.arc = arc
        self.beam = beam

    def run(self):
        tw_arc = self.arc.twiss(self.beam)

        return {
            "table": tw_arc,
            "mux": tw_arc["mux", -1] - tw_arc["mux", 0],
            "muy": tw_arc["muy", -1] - tw_arc["muy", 0],
        }



from .model_madx import LHCMadModel

class LHCXsuiteModel:
    @classmethod
    def from_madxfile(cls, madxfile, sliced=False):
        """
        Create a LHCXsuiteModel from a MAD-X input file
        """
        madmodel=LHCMadModel.from_madxfile(madxfile)

        self=cls.from_madx(madmodel.madx, sliced=sliced, madxfile=madxfile)
        self.madxfile=madxfile
        return self

    @classmethod
    def from_madx(cls, madx, sliced=False):
        import xtrack as xt

        lines = xt.Multiline.from_madx(
            madx=madx, enable_layout_data=True, return_lines=True
        )
        lines["b1"] = lines.pop("lhcb1")
        lines["b2"] = lines.pop("lhcb2")
        lines["b1"].particle_ref = xt.Particles(
            mass0=xt.PROTON_MASS_EV, p0c=450e9
        )
        lines["b2"].particle_ref = xt.Particles(
            mass0=xt.PROTON_MASS_EV, p0c=450e9
        )
        lines["b1"].twiss_default["method"] = "4d"
        lines["b2"].twiss_default["method"] = "4d"
        lines["b1"].twiss_default["co_search_at"] = "ip7"
        lines["b2"].twiss_default["co_search_at"] = "ip7"
        lines["b1"].twiss_default["strengths"] = True
        lines["b2"].twiss_default["strengths"] = True
        lines["b1"].twiss_default["compute_chromatic_properties"] = False
        lines["b2"].twiss_default["compute_chromatic_properties"] = False
        if sliced:
            for ln, ll in list(lines.items()):
                ls = ll.copy()
                ls.slice_thick_elements(
                    slicing_strategies=[
                        xt.Strategy(slicing=None),
                        xt.Strategy(
                            slicing=xt.Uniform(8, mode="thick"), name="mb.*"
                        ),
                        xt.Strategy(
                            slicing=xt.Uniform(8, mode="thick"), name="mq.*"
                        ),
                    ]
                )
                lines[f"{ln}s"] = ls

        lhc = xt.Multiline(lines)
        out = cls(multiline=lhc)
        return out

    @classmethod
    def from_json(cls, jsonfile):
        import xtrack as xt

        lhc = xt.Multiline.from_json(jsonfile)
        return cls(lhc, jsonfile=jsonfile)

    def to_json(self, jsonfile="lhc.json"):
        self.multiline.to_json(jsonfile)

    def __init__(
        self,
        multiline=None,
        optics=None,
        settings=None,
        jsonfile=None,
        madxfile=None,
    ):
        self.multiline = multiline
        self.optics = optics
        self.settings = settings
        self.jsonfile = jsonfile
        self._var_values = multiline._xdeps_vref._owner
        self.vref = multiline._xdeps_vref
        self.mgr = multiline._xdeps_manager
        self.sequence = {1: multiline.b1, 2: multiline.b2}

    @property
    def b1(self):
        return self.multiline.b1

    @property
    def b2(self):
        return self.multiline.b2

    @property
    def b1s(self):
        return self.multiline.b1s

    @property
    def b2s(self):
        return self.multiline.b2s

    def __getitem__(self, key):
        return self._var_values[key]

    def __setitem__(self, key, value):
        self.vref[key] = value

    def update_vars(self, strengths):
        for k, v in strengths.items():
            self[k] = v

    def update_knobs(self, knobs):
        for k, knob in knobs.items():
            self[k] = knob.value
            for wn, value in knob.weights.items():
                name = f"{wn}_from_{k}"
                self[name] = value
                task = self.mgr.tasks.get(self.vref[wn])
                if (
                    task is not None
                    and k in task.dependencies
                    and name not in task.dependencies
                ):
                    raise ValueError(f"{wn} depends on {k} but not on {name}")
                if task is None or (
                    task is not None and k not in task.dependencies
                ):
                    self.vref[wn] += self.vref[name] * self.vref[k]

    def update(self, src):
        if hasattr(src, "strengths"):
            self.update_vars(src.strengths)
        else:
            self.update_vars(src)
        if hasattr(src, "knobs"):
            self.update_knobs(src.knobs)






from .section import (
    LHCSection,
    lhcprev,
    lhcsucc,
    sort_n,
)
from .model_madx import LHCMadModel


class LHCIR(LHCSection):
    """
    Model of an LHC Interaction Region

    twiss_from_init(beam) -> Make twiss using boundary conditions
    twiss_from_params(beam) -> Make twiss using parameters at the IP
    twiss_full(beam) -> Make twiss using full sequence
    """

    knobs = []
    _extra_param_names = []

    @classmethod
    def from_madx(cls, madx, name=None):
        madmodel=LHCMadModel(madx)
        if name is None:
            name = cls.name
        irn = int(name[-1])
        strength_names = []
        strength_names += madmodel.filter(f"kq[xt]?.*[lr]{irn}$")
        quads = madmodel.filter(f"ktq[x].*[lr]{irn}$")
        quads += madmodel.filter(f"kq[t]?[l]?[0-9][0-9]?\..*[lr]{irn}b[12]$")
        # quads += madmodel.filter(madx, f"kq[t]?.*[lr]{irn}$")
        strength_names += sort_n(quads)
        acb = madmodel.filter(f"acbx.*[lr]{irn}$")
        acb += madmodel.filter(f"acb.*[lr]{irn}b[12]$")
        strength_names += sort_n(acb)
        knobs = madmodel.make_and_set0_knobs(cls.knobs)
        strengths = {st: madx.globals[st] for st in strength_names}
        for knob in knobs:
            madx.globals[knob] = knobs[knob].value
        params = {}
        return cls(name, strengths, params, knobs)

    def __init__(
        self,
        name=None,
        strengths=None,
        params=None,
        knobs=None,
        start=None,
        end=None,
    ):
        if name is None:
            name = self.__class__.name
        irn = int(name[-1])
        start = f"s.ds.l{irn}"
        end = f"e.ds.r{irn}"
        super().__init__(name, start, end, strengths, params, knobs)
        self.arc_left_name = f"a{lhcprev(irn)}{irn}"
        self.arc_right_name = f"a{irn}{lhcsucc(irn)}"
        self.init_left = None
        self.init_right = None
        self.irn = irn
        self.startb12 = {1: f"s.ds.l{irn}.b1", 2: f"s.ds.l{irn}.b2"}
        self.endb12 = {1: f"e.ds.r{irn}.b1", 2: f"e.ds.r{irn}.b2"}
        self.param_names=self._get_param_default_names()
        self.param_names.extend(self._extra_param_names)

    @property
    def arc_left(self):
        return getattr(self.parent, self.arc_left_name)

    @property
    def arc_right(self):
        return getattr(self.parent, self.arc_right_name)

    @property
    def quads(self):
        return {k: v for k, v in self.strengths.items() if "kq" in k}

    def set_init(self):
        arcleft = self.arc_left
        self.init_left = {
            1: arcleft.twiss_init(1)[1],
            2: arcleft.twiss_init(2)[1],
        }
        arcright = self.arc_right
        self.init_right = {
            1: arcright.twiss_init(1)[0],
            2: arcright.twiss_init(2)[0],
        }

    def twiss_from_init(self, beam=None):
        if beam is None:
            return [self.twiss_from_init(beam=1), self.twiss_from_init(beam=2)]
        if self.init_left is None:
            self.set_init()
        start = self.startb12[beam]
        end = self.endb12[beam]
        return self.model.sequence[beam].twiss(
            start=start, end=end, init=self.init_left[beam]
        )

    def twiss_from_params(self, beam):
        pass

    def twiss_full(self, beam):
        if beam is None:
            return [self.twiss_full(beam=1), self.twiss_full(beam=2)]
        sequence = self.model.sequence[beam]
        start = self.startb12[beam]
        end = self.endb12[beam]
        init = sequence.twiss().get_twiss_init(start)
        return sequence.twiss(start=start, end=end, init=init)

    def plot(self, beam=None, model="init", figlabel=None):
        if beam is None:
            return [self.plot(beam=1), self.plot(beam=2)]
        else:
            if model == "init":
                mktwiss = self.twiss_from_init
            elif model == "full":
                mktwiss = self.twiss_full
            elif model == "params":
                mktwiss = self.twiss_from_params
            if figlabel is None:
                figlabel = f"{self.name}b{beam}"
            return mktwiss(beam).plot(figlabel=figlabel)

    def _get_param_default_names(self):
        ipname = f"ip{self.irn}"
        params = []
        for param in "betx bety alfx alfy dx dy".split():
            for beam in '12':
                params.append(f"{param}{ipname}{beam}")
        for param in "mux muy".split():
            for beam in '12':
                params.append(f"{param}{ipname}b{beam}")
                params.append(f"{param}{ipname}b{beam}_l")
                params.append(f"{param}{ipname}b{beam}_r")
        return params


    def get_params_from_twiss(self, tw1, tw2):
        ipname = f"ip{self.irn}"
        params = {}
        for param in "betx bety alfx alfy dx dy".split():
            for beam, tw in zip([1, 2], [tw1, tw2]):
                params[f"{param}{ipname}{beam}"] = tw[param, ipname]
        for param in "mux muy".split():
            for beam, tw in zip([1, 2], [tw1, tw2]):
                params[f"{param}{ipname}b{beam}"] = (
                    tw[param, self.endb12[beam]]
                    - tw[param, self.startb12[beam]]
                )
                params[f"{param}{ipname}b{beam}_l"] = (
                    tw[param, ipname] - tw[param, self.startb12[beam]]
                )
                params[f"{param}{ipname}b{beam}_r"] = (
                    tw[param, self.endb12[beam]] - tw[param, ipname]
                )
        return params

    def get_params(self, mode="init"):
        if mode == "init":
            tw1 = self.twiss_from_init(1)
            tw2 = self.twiss_from_init(2)
        elif mode == "full":
            tw1 = self.twiss_full(1)
            tw2 = self.twiss_full(2)
        return self.get_params_from_twiss(tw1, tw2)



import json


def madname_from_pcname(pc):
    name = ".".join(pc.split(".")[2:]).lower()
    if name.startswith("rcb"):
        return "a" + name[1:]
    else:
        return "k" + name[1:]

def get_calib_mad(lsa, madname):
    print(madname)
    pcname = lsa.findPCNameByMadStrength(madname)
    if pcname is not None:
        if pcname.split(".")[0] in ["RQD", "RQF", "RB"]:
            pcname = pcname + "B1"
        cal = lsa.getCalibration(pcname)
        if cal is not None:
            return [cal.field, cal.current]

class LHCCircuits:
    @classmethod
    def from_lsa(cls):
        import pjlsa
        lsa=pjlsa.LSAClient()

    @classmethod
    def from_json(cls, filename):
        pass

    def __init__(self, circuits):
        self.circuits = circuits

    def to_json(self, filename):
        pass


class LHCCircuit:
    def __init__(
        self,
        name,
        madname,
        calib,
        r1,
        l1,
        c1,
        imax,
        imin,
        vmax,
        vmin,
        ipmax,
        ipmin,
        vpmax,
        vpmin,
        polarity,
    ):
        self.name = name
        self.madname = madname
        self.calib = calib
        self.r1 = r1
        self.l1 = l1
        self.c1 = c1
        self.imax = imax
        self.imin = imin
        self.vmax = vmax
        self.vmin = vmin
        self.ipmax = ipmax
        self.ipmin = ipmin
        self.vpmax = vpmax
        self.vpmin = vpmin
        self.polarity = polarity


class LHCCircuit2in1:
    def __init__(self, circuit1, circuit2, rc):
        self.circuit1 = circuit1
        self.circuit2 = circuit2
        self.rc = rc



from .irs import LHCIR


class LHCIR2(LHCIR):
    name = "ir2"
    knobs = [
        "on_x2h",
        "on_sep2h",
        "on_x2v",
        "on_sep2v",
        "on_a2",
        "on_oh2",
        "on_xip2b2",
        "on_xip2b1",
        "on_o2",
        "on_ov2",
        "on_yip2b1",
        "on_yip2b2",
    ]



from .irs import LHCIR


class LHCIR8(LHCIR):
    name = "ir8"
    knobs = [
        "on_x8h",
        "on_sep8h",
        "on_x8v",
        "on_sep8v",
        "on_o8",
        "on_oh8",
        "on_xip8b2",
        "on_xip8b1",
        "on_a8",
        "on_ov8",
        "on_yip8b1",
        "on_yip8b2",
    ]



"""
LHCModel -> Compute twiss, aperture, match
LHCOptics -> Contains complete optics
LHCIROptics -> Contains strengths, constraints, knobs
LHCArcOptics -> Contains strengths, constraints, knobs
"""

__version__ = "0.0.0"
from .optics import LHCOptics
from .model_xsuite import LHCXsuiteModel, LHCMadModel
from .circuits import LHCCircuits
